package flow

//
// Find a way to improve this mess, maybe it can be merged in one func
//
//

import (
	"errors"
	"fmt"
	"reflect"
	"runtime/debug"
	"sync"
)

type executorFunc func(*Session, ...Data) (Data, error)

// Operation interface
type Operation interface { // Id perhaps?
	//ID() string
	//Set(input Data) // Special var method
	Process(params ...Data) (Data, error)
}

type operation struct {
	sync.Mutex
	flow *Flow
	name string
	kind string

	// Could be a simple ID for operation, but it will depend on flow
	inputs []*operation // still figuring, might be Operation

	fn       interface{}  // the registry func
	executor executorFunc // the executor?
	//processor func(OpCtx, params ...Data) (Data, error)
}

// NewOperation creates an operation
func (f *Flow) newOperation(kind string, inputs []*operation) *operation {
	return &operation{
		Mutex:  sync.Mutex{},
		flow:   f,
		kind:   kind,
		inputs: inputs,
		//name:   fmt.Sprintf("(var)<%s>", name),
	}

}

// Process params are the global inputs
func (o *operation) Process(ginputs ...Data) (Data, error) {
	// Create CTX
	s := o.flow.NewSession()
	return s.Run(o, ginputs...)
}

// make Executor for func
// safe run a func
func (f *Flow) asTrigger(op *operation) executorFunc {
	return func(sess *Session, params ...Data) (Data, error) {
		f.hooks.start(op)

		//panic recoverer, since nodes are not our functions
		var err error
		var res Data
		// Safe thing
		func() {
			defer func() {
				if r := recover(); r != nil {
					debug.PrintStack()
					err = fmt.Errorf("%v", r)
				}
			}()
			res, err = op.executor(sess, params...)
		}()

		if err != nil {
			f.hooks.error(op, err)
		} else {
			f.hooks.finish(op, res)
		}
		return res, err

	}
}

// processInputs will run a list of operations and return reflect values
// to be processed next
// NEW PARALLEL PROCESSING

// Var create a operation
func (f *Flow) Var(name string, initial Data) Operation {
	inputs := f.makeInputs(initial)

	op := f.newOperation("var", inputs)
	op.name = fmt.Sprintf("(var)<%s>", name)

	op.executor = func(sess *Session, ginputs ...Data) (Data, error) {
		val, ok := f.Data[name]
		if !ok {
			var initial Data
			f.hooks.wait(op)
			res, err := sess.RunList(op.inputs, ginputs...)
			if err != nil {
				return nil, err
			}
			if len(res) > 0 {
				initial = res[0]
			}

			val = initial
			f.Data[name] = val
		}
		return val, nil
	}
	return op
}

// SetVar a variable from operation/constant
func (f *Flow) SetVar(name string, data Data) Operation {
	inputs := f.makeInputs(data)
	op := f.newOperation("setvar", inputs)
	op.name = fmt.Sprintf("(setvar)<%s>", name)
	op.executor = func(sess *Session, ginputs ...Data) (Data, error) {
		f.hooks.wait(op)
		res, err := sess.RunList(op.inputs, ginputs...)
		if err != nil {
			return nil, err
		}
		f.Data[name] = res[0]
		return res[0], nil
	}

	return op
}

// Op operation from registry
func (f *Flow) Op(name string, params ...interface{}) Operation {
	inputs := f.makeInputs(params...)

	// Grab executor here
	registryFn, err := f.registry.Get(name)
	if err != nil {
		return f.ErrOp(err)
	}
	op := f.newOperation("func", inputs)
	op.name = name
	// make executor from registry func
	op.executor = makeExecutor(op, registryFn)
	f.operations = append(f.operations, op)
	return op
}

// ErrOp define a nil operation that will return error
// Usefull for builders
func (f *Flow) ErrOp(err error) Operation {
	op := f.newOperation("error", nil)
	op.executor = func(*Session, ...Data) (Data, error) { return nil, err }
	return op
}

// Const define a const by defined ID
func (f *Flow) Const(value Data) Operation {
	// Optimize this definition
	constID := -1
	for k, v := range f.consts {
		if reflect.DeepEqual(v, value) {
			constID = k
			break
		}
	}
	if constID == -1 {
		constID = len(f.consts)
		f.consts = append(f.consts, value)
	}

	op := f.newOperation("const", nil)
	op.name = fmt.Sprintf("(const)<%v:%v>", constID, value)
	op.executor = func(*Session, ...Data) (Data, error) { return f.consts[constID], nil }
	//f.operations = append(f.operations, op)

	return op
}

// In define input operation
func (f *Flow) In(paramID int) Operation {

	op := f.newOperation("in", nil)
	op.name = fmt.Sprintf("(in)<%d>", paramID)
	op.executor = func(sess *Session, ginputs ...Data) (Data, error) {
		if paramID < 0 || paramID >= len(ginputs) {
			return nil, ErrInput
		}
		return ginputs[paramID], nil
	}
	//f.operations = append(f.operations, op)
	return op
}

func (f *Flow) makeInputs(params ...Data) []*operation {
	inputs := make([]*operation, len(params))
	for i, p := range params {
		switch v := p.(type) {
		case *operation:
			inputs[i] = v
		default:
			c := f.Const(v)
			inputs[i], _ = c.(*operation)
		}
	}
	return inputs
}

// make any go func as an executor
// Trigger
func makeExecutor(op *operation, fn interface{}) executorFunc {
	// ExecutorFunc
	return func(sess *Session, ginput ...Data) (Data, error) {
		// Change to wait to wait for the inputs

		op.flow.hooks.wait(op)
		inRes, err := sess.RunList(op.inputs, ginput...)
		if err != nil {
			return nil, err
		}

		fnval := reflect.ValueOf(fn)
		callParam := make([]reflect.Value, len(inRes))
		for i, r := range inRes {
			if r == nil {
				callParam[i] = reflect.Zero(fnval.Type().In(i))
			} else {
				callParam[i] = reflect.ValueOf(r)
			}
		}

		// Start again and execute function
		op.flow.hooks.start(op)
		fnret := fnval.Call(callParam)
		if len(fnret) == 0 {
			return nil, nil
		}
		// Output erroring
		if len(fnret) > 1 && (fnret[len(fnret)-1].Interface() != nil) {
			err, ok := fnret[len(fnret)-1].Interface().(error)
			if !ok {
				err = errors.New("unknown error")
			}
			return nil, err
		}

		// THE RESULT
		ret := fnret[0].Interface()
		return ret, nil
	}
}
