package flow

import (
	"errors"
	"sync"
)

// Session operation session
type Session struct {
	*sync.Map
	flow *Flow
}

// NewSession creates a running context
func (f *Flow) NewSession() *Session {
	return &Session{
		Map:  &sync.Map{},
		flow: f,
	}
}

// Run run operation
func (s *Session) Run(op *operation, ginputs ...Data) (Data, error) {
	op.Lock()
	defer op.Unlock()
	// Load from cache if any
	if v, ok := s.Load(op); ok {
		return v, nil
	}

	res, err := s.flow.asTrigger(op)(s, ginputs...)
	if err != nil {
		return nil, err
	}

	s.Store(op, res)
	return res, nil

}

// RunList more than one operations in this session
func (s *Session) RunList(ops []*operation, ginputs ...Data) ([]Data, error) {
	nOps := len(ops)

	// Total inputs
	callParam := make([]Data, nOps)

	callErrors := ""
	paramMutex := sync.Mutex{}
	// Parallel processing if inputs
	wg := sync.WaitGroup{}
	wg.Add(nOps)
	for i, op := range ops {
		go func(i int, op *operation) {
			defer wg.Done()

			res, err := s.Run(op, ginputs...)
			paramMutex.Lock()
			defer paramMutex.Unlock()
			if err != nil {
				callErrors += err.Error() + "\n"
				return
			}
			callParam[i] = res
		}(i, op)
	}
	wg.Wait()

	if callErrors != "" {
		return nil, errors.New(callErrors)
	}

	return callParam, nil
}
